/**
 * BOINC Curses
 * http://www.oook.cz/bsd/boinc_curses/
 * Copyright (C) 2006 Jan Hornyak <pav@oook.cz>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * To view the GNU Lesser General Public License visit
 * http://www.gnu.org/copyleft/lesser.html
 * or write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <curses.h>
#include <signal.h>
#include <sys/ioctl.h>

#include <gui_rpc_client.h>
#include <version.h>
#include <util.h>
#if BOINC_MAJOR_VERSION > 5 || (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION > 5)
#include <common_defs.h>
#endif

#define BOINC_CURSES_VERSION "0.1.7"
#define PAINT_RESULTS 1
#define PAINT_TRANSFERS 2
#define REFRESH_RATE 10

static void finish(int signal);
static void handle_winch(int signal);
static void reconnect(int signal);
static void prettyCPUTime(char *output, double time);
static void paintTable(char bufs[][64], attr_t attrs[], short colors[], int count, int colcount, int &lineno, int painttype);
static void projectOp(const char* message, const char* op);
static int getIndex(int count, const char* message);
static void paintHelp();
static void printUsage();

RPC_CLIENT rpc;
char passkey[256] = "";

int main(int argc, char **argv) {

	int i;

	char dbdir[1024] = "";
	char *buf = (char *)malloc(1024);
	char *buf2 = (char *)malloc(1024);

	/* getopt */
	while ((i = getopt(argc, argv, "d:hp:v")) != -1) {
		switch (i) {
			case 'd':
				strncpy(dbdir, optarg, sizeof(dbdir) - 1);
				dbdir[sizeof(dbdir) - 1] = '\0';
				break;
			case 'h':
				printUsage();
				break;
			case 'p':
				strncpy(passkey, optarg, sizeof(passkey) - 1);
				dbdir[sizeof(passkey) - 1] = '\0';
				break;
			case 'v':
				printf("boinc_curses %s\n", BOINC_CURSES_VERSION);
				exit(0);
				break;
			default:
				printUsage();
		}
	}


	/* authenticate */
	if (passkey[0] == '\0') {
		if (dbdir[0] == '\0')
			sprintf(buf, "%s/gui_rpc_auth.cfg", BOINC_HOME);
		else
			sprintf(buf, "%s/gui_rpc_auth.cfg", dbdir);
		FILE *f = fopen(buf, "r");
		if (!f) {
			printf("ERROR: can't read %s\n", buf);
			exit(3);
		}
		fgets(passkey, 256, f);
		fclose(f);
	}

	/* set up signal handlers */
	signal(SIGINT, finish);
	signal(SIGTERM, finish);
	signal(SIGWINCH, handle_winch);
	signal(SIGPIPE, reconnect);

	/* init ncurses */
	initscr();
	nonl();
	cbreak();
	halfdelay(REFRESH_RATE);
	noecho();

	start_color();
	for (i = 0; i < 8; i++) init_pair(i, i, COLOR_BLACK);

	refresh();

	/* connect to core */
	reconnect(0);

	/* loop */
	int lineno, colcount;
	int x, y, storey;
	int count;
	int seqno = 0;

	for (;;) {

		erase();
		lineno = 0;

		/* get state */
		CC_STATE state;
		rpc.get_state(state);

		/* get projects */
		PROJECTS projects;
		PROJECT* project;
		rpc.get_project_status(projects);

		/* get results */
		RESULTS results;
		rpc.get_results(results);
		count = results.results.size();
		mvprintw(lineno, 0, "%d results", count);
		getyx(stdscr, x, y);
		mvchgat(lineno, 0, -1, A_BOLD, 0, NULL);

		/* get cc status */
#if BOINC_MAJOR_VERSION > 5 || (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION > 5)
		CC_STATUS cc_status;
		rpc.get_cc_status(cc_status);
#else
		struct {
			int network_mode;
			int task_mode;
		} cc_status;
		rpc.get_network_mode(cc_status.network_mode);
		rpc.get_run_mode(cc_status.task_mode);
#endif

		/* project settings */
		move(0, y);
		storey = y;
		count = projects.projects.size();
		if (count > 0) {
			for (i = 0; i < count; i++) {
				int need_dash = 0;
				project = projects.projects.at(i);
				if (project->suspended_via_gui || project->dont_request_more_work) printw("  %s:", project->project_name.c_str());
				if (project->suspended_via_gui) {
					printw(" suspended");
					need_dash = 1;
				}
				if (project->dont_request_more_work) {
					if (need_dash) printw(",");
					printw(" no more work");
				}
			}
		}
		getyx(stdscr, x, y);
		mvchgat(lineno, storey, -1, A_BOLD, COLOR_RED, NULL);
		storey = y;

		/* display abnormal client states */
		if (cc_status.network_mode == RUN_MODE_NEVER || cc_status.task_mode == RUN_MODE_NEVER) {
			move(0, y);
			if (cc_status.network_mode == RUN_MODE_NEVER) {
				printw("  Network disabled");
			}
			if (cc_status.task_mode == RUN_MODE_NEVER) {
				printw("  Computation suspended");
			}
		}
		mvchgat(lineno, storey, -1, A_BOLD, COLOR_MAGENTA, NULL);
		lineno++;

		/* list results */
		count = results.results.size();
		if (count > 0) {
			RESULT* result;
			colcount = 7;
			char bufs[colcount * count][64];
			attr_t attrs[count];
			short colors[count];
			for (i = 0; i < count; i++) {
				result = results.results.at(i);
				RESULT* state_result = state.lookup_result(result->project_url, result->name);
				if (state_result && state_result->project) {
					strcpy(bufs[(i * colcount) + 0], state_result->project->project_name.c_str());
					sprintf(bufs[(i * colcount) + 1], "%s %d", state_result->wup->avp->app_name.c_str(), state_result->wup->avp->version_num);
				} else {
					strcpy(bufs[(i * colcount) + 0], "???");
					strcpy(bufs[(i * colcount) + 1], "???");
				}
				strcpy(bufs[(i * colcount) + 2], result->name.c_str());
				switch (result->state) {
					case RESULT_NEW:
						strcpy(buf, "New");
						break;
					case RESULT_FILES_DOWNLOADING:
						if (result->ready_to_report) strcpy(buf, "Download failed");
						else strcpy(buf, "Downloading");
						break;
					case RESULT_FILES_DOWNLOADED:
						if (result->project_suspended_via_gui) strcpy(buf, "Project suspended");
#if BOINC_MAJOR_VERSION < 5 || (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION < 5)
						else if (result->aborted_via_gui) strcpy(buf, "Aborted by user");
#endif
						else if (result->suspended_via_gui) strcpy(buf, "Suspended");
						else if (result->active_task) {
							if (cc_status.task_mode == RUN_MODE_NEVER) strcpy(buf, "Computation suspended");
							else if (result->scheduler_state == CPU_SCHED_SCHEDULED) strcpy(buf, "Running");
							else if (result->scheduler_state == CPU_SCHED_PREEMPTED) strcpy(buf, "Preempted");
							else strcpy(buf, "Ready to run");
						} else strcpy(buf, "Ready to run");
						break;
					case RESULT_COMPUTE_ERROR:
						strcpy(buf, "Failed");
						break;
					case RESULT_FILES_UPLOADING:
						if (result->ready_to_report) strcpy(buf, "Upload failed");
						else strcpy(buf, "Uploading");
						break;
					case RESULT_FILES_UPLOADED:
						if (result->got_server_ack) strcpy(buf, "Acknowledged");
						else strcpy(buf, "Ready to report");
						break;
#if (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION >= 5) || BOINC_MAJOR_VERSION > 5
					case RESULT_ABORTED:
						strcpy(buf, "Aborted");
						break;
#endif
					default:
						sprintf(buf, "State %d", result->state);
				}
				strcpy(bufs[(i * colcount) + 3], buf);
				if (result->active_task && result->scheduler_state == CPU_SCHED_SCHEDULED) {
					prettyCPUTime(buf, result->current_cpu_time);
				} else {
					prettyCPUTime(buf, result->final_cpu_time);
				}
				sprintf(bufs[(i * colcount) + 4], "%s", buf);
				if (result->active_task) {
					sprintf(bufs[(i * colcount) + 5], "%.3f%%", 100 * result->fraction_done);
				} else {
					strcpy(bufs[(i * colcount) + 5], "");
				}
				prettyCPUTime(buf, result->estimated_cpu_time_remaining);
				sprintf(bufs[(i * colcount) + 6], "%s", buf);
				int age = result->report_deadline - time(NULL);
				if (age <= 86400) {
					if (age < 0) {
						strcat(bufs[(i * colcount) + 6], ", deadline missed!");
					} else {
						prettyCPUTime(buf, age);
						strcat(bufs[(i * colcount) + 6], ", deadline in ");
						strcat(bufs[(i * colcount) + 6], buf);
					}
				}
				/* Color the line */
				if (result->active_task && result->scheduler_state == CPU_SCHED_SCHEDULED && cc_status.task_mode != RUN_MODE_NEVER) {
					attrs[i] = A_BOLD; colors[i] = COLOR_WHITE;
				} else if (result->state == RESULT_FILES_DOWNLOADING || result->state == RESULT_FILES_UPLOADING) {
					attrs[i] = A_NORMAL; colors[i] = COLOR_YELLOW;
				} else if (result->state == RESULT_FILES_UPLOADED) {
					attrs[i] = A_NORMAL; colors[i] = COLOR_GREEN;
				} else {
					attrs[i] = A_NORMAL; colors[i] = COLOR_WHITE;
				}
			}

			paintTable(bufs, attrs, colors, count, colcount, lineno, PAINT_RESULTS);
		}

		/* get file transfers */
		FILE_TRANSFERS transfers;
		rpc.get_file_transfers(transfers);
		count = transfers.file_transfers.size();
		if (count > 0) {
			mvprintw(lineno, 0, "%d file transfers", count);
			mvchgat(lineno, 0, -1, A_BOLD, 0, NULL);
			lineno++;
			FILE_TRANSFER* transfer;
			colcount = 5;
			char bufs[colcount * count][64];
			attr_t attrs[count];
			short colors[count];
			for (i = 0; i < count; i++) {
				transfer = transfers.file_transfers.at(i);
				strcpy(bufs[(i * colcount) + 0], transfer->project_name.c_str());
				strcpy(bufs[(i * colcount) + 1], transfer->name.c_str());
				if (transfer->generated_locally) strcpy(bufs[(i * colcount) + 2], "up");
				else strcpy(bufs[(i * colcount) + 2], "down");
				sprintf(bufs[(i * colcount) + 3], "%d kB", (int)(transfer->nbytes/1024));
				if (transfer->xfer_active) {
					sprintf(bufs[(i * colcount) + 4], "%d kB", (int)((transfer->nbytes - transfer->bytes_xferred)/1024));
				} else if (transfer->num_retries > 0) {
					sprintf(bufs[(i * colcount) + 4], "%d fail", transfer->num_retries);
					if (transfer->num_retries > 1) strcat(bufs[(i * colcount) + 4], "s");
					if (transfer->next_request_time > time(NULL)) {
						prettyCPUTime(buf, transfer->next_request_time - time(NULL));
						sprintf(buf2, ", next try in %s", buf);
						strcat(bufs[(i * colcount) + 4], buf2);
					}
				} else {
					strcpy(bufs[(i * colcount) + 4], "");
				}
				/* Color the line */
				if (transfer->xfer_active) {
					attrs[i] = A_BOLD; colors[i] = COLOR_WHITE;
				} else {
					attrs[i] = A_NORMAL; colors[i] = COLOR_WHITE;
				}
			}

			paintTable(bufs, attrs, colors, count, colcount, lineno, PAINT_TRANSFERS);
		}

		/* messages */
		MESSAGES messages;
		rpc.get_messages(seqno, messages);
		count = messages.messages.size();
		if (count > 0) {
			MESSAGE* message;
			for (i = count-1; i >= 0; i--) {
				message = messages.messages.at(i);
				int age = time(NULL) - message->timestamp;
				if (age >= 600) {
					/* old message, skip and skip next time */
					seqno = message->seqno;
					break;
				} else {
					if (i == count - 1) {
						mvprintw(lineno, 0, "Recent messages");
						mvchgat(lineno, 0, -1, A_BOLD, 0, NULL);
						lineno++;
					}
					time_t clock = message->timestamp;
					strftime(buf, 9, "%k:%M:%S", localtime(&clock));
					if (message->project.size()) {
						mvprintw(lineno, 0, "%8s  %s  %s", buf, message->project.c_str(), message->body.c_str());
					} else {
						mvprintw(lineno, 0, "%8s  %s", buf, message->body.c_str());
					}
					/* Color the line */
					if (message->priority > 1) mvchgat(lineno, 0, -1, A_BOLD, COLOR_RED, NULL);
					else if (age < 5) mvchgat(lineno, 0, -1, A_BOLD, 0, NULL);
					else if (age < 60) mvchgat(lineno, 0, -1, A_BOLD, COLOR_WHITE, NULL);
					else mvchgat(lineno, 0, -1, A_NORMAL, COLOR_WHITE, NULL);
					lineno++;
				}
			}
		}

		/* sweep cursor */
		move(LINES-1, COLS-1);

		refresh();

		/* keyboard */
		int c = getch();

		switch (c) {
			case 'a':
				/* allow more work */
				projectOp("ALLOW MORE WORK: ", "allowmorework");
				break;
			case 'A':
				/* abort result */
				count = results.results.size();
				if ((i = getIndex(count, "ABORT WORK UNIT")) != -1) {
					RESULT* result;
					result = results.results.at(i);
					rpc.result_op(*result, "abort");
				}
				break;
			case 'b':
				/* run benchmarks */
				rpc.run_benchmarks();
				break;
			case 'C':
				/* toggle computation mode */
#if BOINC_MAJOR_VERSION > 5 || (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION > 6)
				if (cc_status.task_mode == RUN_MODE_NEVER) rpc.set_run_mode(RUN_MODE_AUTO, 60);
				else rpc.set_run_mode(RUN_MODE_NEVER, 60);
#else
				if (cc_status.task_mode == RUN_MODE_NEVER) rpc.set_run_mode(RUN_MODE_AUTO);
				else rpc.set_run_mode(RUN_MODE_NEVER);
#endif
				break;
			case 'h':
			case '?':
				/* show help */
				paintHelp();
				break;
			case 'n':
				/* don't allow more work */
				projectOp("NO MORE WORK: ", "nomorework");
				break;
			case 'N':
				/* toggle network mode */
#if BOINC_MAJOR_VERSION > 5 || (BOINC_MAJOR_VERSION == 5 && BOINC_MINOR_VERSION > 6)
				if (cc_status.network_mode == RUN_MODE_NEVER) rpc.set_network_mode(RUN_MODE_AUTO, 60);
				else rpc.set_network_mode(RUN_MODE_NEVER, 60);
#else
				if (cc_status.network_mode == RUN_MODE_NEVER) rpc.set_network_mode(RUN_MODE_AUTO);
				else rpc.set_network_mode(RUN_MODE_NEVER);
#endif
				break;
			case 'q':
				/* quit */
				finish(0);
				break;
			case 'r':
				/* resume result */
				count = results.results.size();
				if ((i = getIndex(count, "RESUME WORK UNIT")) != -1) {
					RESULT* result;
					result = results.results.at(i);
					rpc.result_op(*result, "resume");
				}
				break;
			case 'R':
				/* resume project */
				projectOp("RESUME PROJECT: ", "resume");
				break;
			case 's':
				/* suspend result */
				count = results.results.size();
				if ((i = getIndex(count, "SUSPEND WORK UNIT")) != -1) {
					RESULT* result;
					result = results.results.at(i);
					rpc.result_op(*result, "suspend");
				}
				break;
			case 'S':
				/* suspend project */
				projectOp("SUSPEND PROJECT: ", "suspend");
				break;
			case 't':
				/* retry transfer */
				count = transfers.file_transfers.size();
				if ((i = getIndex(count, "RETRY TRANSFER")) != -1) {
					FILE_TRANSFER* transfer;
					transfer = transfers.file_transfers.at(i);
					rpc.file_transfer_op(*transfer, "retry");
				}
				break;
			case 'u':
				/* update project */
				projectOp("UPDATE PROJECT: ", "update");
				break;
		}
	}

	finish(0);
}

static void finish(int signal) {

	endwin();

	exit(0);
}

static void handle_winch(int signal) {

	endwin();
	refresh();
}

static void reconnect(int signal) {
	int rpcerr;

	for(;;) {
		/* connect to core */
		rpcerr = rpc.init("localhost", GUI_RPC_PORT);
		if (rpcerr == 0)
			break;
		if (signal == 0) {
			erase();
			mvprintw(0, 0, "ERROR: can't connect to boinc-client (%s)\n", boincerror(rpcerr));
			refresh();
			exit(1);
		}
		erase();
		mvprintw(0, 0, "WARNING: can't connect to boinc-client (%s)\n", boincerror(rpcerr));
		refresh();
		sleep(1);
	}

	rpcerr = rpc.authorize(passkey);
	if (rpcerr != 0) {
		endwin();
		printf("ERROR: failed to authorize (%s)\n", boincerror(rpcerr));
		exit(2);
	}
	erase();
	refresh();
}

static void prettyCPUTime(char *output, double time) {

	int hours, mins, secs;

	int itime = (int)time;
	secs = itime % 60;
	itime -= secs;
	itime /= 60;
	mins = itime % 60;
	itime -= mins;
	itime /= 60;
	hours = itime;

	char buf[4];

	if (hours > 0) sprintf(output, "%d:", hours);
	else strcpy(output, "");
	if (hours > 0 || mins > 0) {
		if (hours > 0) sprintf(buf, "%02d:", mins);
		else sprintf(buf, "%d:", mins);
		strcat(output, buf);
	} else strcat(output, "");
	if (hours > 0 || mins > 0 || secs > 0) {
		if (hours > 0 || mins > 0) sprintf(buf, "%02d", secs);
		else sprintf(buf, "%d", secs);
		strcat(output, buf);
	} else strcat(output, "");

}

static void paintTable(char bufs[][64], attr_t attrs[], short colors[], int count, int colcount, int &lineno, int painttype) {

	/* determine the width of columns */
	int i, j;
	size_t bufmaxlen[colcount], skip[colcount];
	for (j = 0; j < colcount; j++) bufmaxlen[j] = skip[j] = 0;
	for (i = 0; i < count; i++) {
		for (j = 0; j < colcount; j++) {
			if (strlen(bufs[(i * colcount) + j]) > bufmaxlen[j]) bufmaxlen[j] = strlen(bufs[(i * colcount) + j]);
		}
	}

	/* determine the width of table */
	int space, height, width;
	int totallen = 0;
	for (j = 0; j < colcount; j++) {
		totallen += bufmaxlen[j];
	}
	/* narrow spaces on small screen */
	getmaxyx(stdscr, height, width);
	if ((totallen + ((colcount - 1) * 2)) > width) {
		space = 1;
	} else {
		space = 2;
	}
	totallen += (colcount - 1) * space;
	/* decide to skip some columns */
	if (painttype == PAINT_RESULTS) {
		/* skip project name */
		if (totallen > width) {
			skip[0] = 1;
			totallen -= bufmaxlen[0];
			totallen -= space;
		}
		/* trim wu name */
		if (totallen > width) {
			int j = 2;
			int oversize = totallen - width;
			int targetlen = bufmaxlen[j] - oversize;
			if (targetlen < 5) targetlen = 5;
			char formatter[10];
			sprintf(formatter, "%%.%ds'", targetlen - 1);
			for (i = 0; i < count; i++) {
				sprintf(bufs[(i * colcount) + j], formatter, bufs[(i * colcount) + j]);
			}
			bufmaxlen[j] = targetlen;
		}
	}

	/* paint the table */
	for (i = 0; i < count; i++) {
		int colno = 0;
		char formatter[4];
		for (j = 0; j < colcount; j++) {
			if (skip[j]) continue;
			strcpy(formatter, "%s");
			if (painttype == PAINT_RESULTS) {
				switch (j) {
					case 4: case 5: case 6:
						sprintf(formatter, "%%%zds", bufmaxlen[j]);
					break;
				}
			}
			if (painttype == PAINT_TRANSFERS) {
				switch (j) {
					case 3: case 4:
						sprintf(formatter, "%%%zds", bufmaxlen[j]);
					break;
				}
			}
			mvprintw(lineno, colno, formatter, bufs[(i * colcount) + j]);
			colno = colno + bufmaxlen[j] + space;
		}
		mvchgat(lineno, 0, -1, attrs[i], colors[i], NULL);
		lineno++;
	}
	lineno++;
}

static void projectOp(const char* message, const char* op) {
	int c, i, count;
	PROJECTS projects;
	PROJECT* project;
	rpc.get_project_status(projects);
	count = projects.projects.size();
	mvprintw(LINES - 1, 0, message);
	for (i = 0; i < count; i++) {
		project = projects.projects.at(i);
		printw("[ %d - %s ] ", i+1, project->project_name.c_str());
	}
	refresh();
	cbreak();
	c = getch();
	if (c >= '1' && c <= '9') {
		i = c - '1';
		if (i < count) {
			/* project_op update */
			project = projects.projects.at(i);
			rpc.project_op(*project, op);
		}
	}
	halfdelay(REFRESH_RATE);
}

static int getIndex(int count, const char* message) {

	int c;
	int i = 0;

	if (count > 0) {
		/* Clip the count at two digits */
		if (count > 99) count = 99;

		mvprintw(LINES - 1, 0, "%s: ", message);
		if (count >= 10) {
			printw("01 - %d", count);
		} else {
			printw("1");
			if (count > 1) printw(" - %d", count);
		}
		printw(" ? ");
		refresh();
		cbreak();
		if (count >= 10) {
			c = getch();
			if (c >= '0' && c <= '9') {
				i = (c - '0') * 10;
			}
			printw("%c", c);
			refresh();
		}
		c = getch();
		if (c >= '0' && c <= '9') {
			i += (c - '0');
		}
		halfdelay(REFRESH_RATE);
	}

	if (i > count) i = 0;
	if (i < 0) i = 0;

	/* Convert to array index */
	i--;

	return i;
}

static void paintHelp() {

	cbreak();
	erase();
	mvprintw(0, 0, "boinc_curses %s", BOINC_CURSES_VERSION);
	mvprintw(2, 0, "KEY BINDINGS");
	mvprintw(3, 0, "a - allow more work");
	mvprintw(3, 40, "A - abort result");
	mvprintw(4, 0, "b - run benchmarks");
	mvprintw(5, 40, "C - toggle run mode");
	mvprintw(6, 0, "h - show help");
	mvprintw(7, 0, "n - set no more work");
	mvprintw(7, 40, "N - toggle network mode");
	mvprintw(8, 0, "q - quit");
	mvprintw(9, 0, "r - resume result");
	mvprintw(9, 40, "R - resume project");
	mvprintw(10, 0, "s - suspend result");
	mvprintw(10, 40, "S - suspend project");
	mvprintw(11, 0, "t - retry transfer");
	mvprintw(12, 0, "u - update project");
	mvprintw(14, 0, "Hit any key to continue");
	refresh();
	getch();
	halfdelay(REFRESH_RATE);

}

static void printUsage() {

	printf("boinc_curses %s\n", BOINC_CURSES_VERSION);
	printf("http://www.oook.cz/bsd/boinc_curses/\n\n");
	printf("-d dir       directory containing gui_rpc_auth.cfg\n");
	printf("-h           print this help\n");
	printf("-p passkey   passkey to authenticate to boinc-client\n");
	printf("-v           show version\n");

	exit(0);

}
